@@ -319,14 +319,14 @@ inline std::string Demangle(char const *msg_str) {
   return string(msg_str);
 }
 
-inline std::string StackTrace() {
+inline std::string StackTrace(
+    const size_t stack_size = DMLC_LOG_STACK_TRACE_SIZE) {
   using std::string;
   std::ostringstream stacktrace_os;
-  const int MAX_STACK_SIZE = DMLC_LOG_STACK_TRACE_SIZE;
-  void *stack[MAX_STACK_SIZE];
-  int nframes = backtrace(stack, MAX_STACK_SIZE);
+  std::vector<void*> stack(stack_size);
+  int nframes = backtrace(stack.data(), static_cast<int>(stack_size));
   stacktrace_os << "Stack trace returned " << nframes << " entries:" << std::endl;
-  char **msgs = backtrace_symbols(stack, nframes);
+  char **msgs = backtrace_symbols(stack.data(), nframes);
   if (msgs != nullptr) {
     for (int frameno = 0; frameno < nframes; ++frameno) {
       string msg = dmlc::Demangle(msgs[frameno]);
@@ -344,7 +344,7 @@ inline std::string demangle(char const* msg_str) {
   return std::string();
 }
 
-inline std::string StackTrace() {
+inline std::string StackTrace(const size_t stack_size = 0) {
   return std::string("stack traces not available when "
   "DMLC_LOG_STACK_TRACE is disabled at compile time.");
 }
