@@ -33,7 +33,11 @@ AllreduceRobust::AllreduceRobust(void) {
 }
 void AllreduceRobust::Init(int argc, char* argv[]) {
   AllreduceBase::Init(argc, argv);
-  result_buffer_round = std::max(world_size / num_global_replica, 1);
+  if (num_global_replica == 0) {
+    result_buffer_round = -1;
+  } else {
+    result_buffer_round = std::max(world_size / num_global_replica, 1);
+  }
 }
 /*! \brief shutdown the engine */
 void AllreduceRobust::Shutdown(void) {
@@ -46,6 +50,12 @@ void AllreduceRobust::Shutdown(void) {
   // execute check ack step, load happens here
   utils::Assert(RecoverExec(NULL, 0, ActionSummary::kCheckAck, ActionSummary::kSpecialOp),
                 "Shutdown: check ack must return true");
+#ifdef __APPLE__
+  // In OSX, one worker shutdowns and closes sockets while rest still run kCheckAck
+  // This cause rest workers checkandrecover and hang inf, https://github.com/dmlc/xgboost/pull/3818
+  // TODO: a fundamental fix for this
+  sleep(2);
+#endif
   AllreduceBase::Shutdown();
 }
 /*!
@@ -86,7 +96,8 @@ void AllreduceRobust::Allreduce(void *sendrecvbuf_,
   bool recovered = RecoverExec(sendrecvbuf_, type_nbytes * count, 0, seq_counter);
   // now we are free to remove the last result, if any
   if (resbuf.LastSeqNo() != -1 &&
-      (resbuf.LastSeqNo() % result_buffer_round != rank % result_buffer_round)) {
+      (result_buffer_round == -1 ||
+       resbuf.LastSeqNo() % result_buffer_round != rank % result_buffer_round)) {
     resbuf.DropLast();
   }
   if (!recovered && prepare_fun != NULL) prepare_fun(prepare_arg);
@@ -118,7 +129,8 @@ void AllreduceRobust::Broadcast(void *sendrecvbuf_, size_t total_size, int root)
   bool recovered = RecoverExec(sendrecvbuf_, total_size, 0, seq_counter);
   // now we are free to remove the last result, if any
   if (resbuf.LastSeqNo() != -1 &&
-      (resbuf.LastSeqNo() % result_buffer_round != rank % result_buffer_round)) {
+      (result_buffer_round == -1 ||
+       resbuf.LastSeqNo() % result_buffer_round != rank % result_buffer_round)) {
     resbuf.DropLast();
   }
   void *temp = resbuf.AllocTemp(1, total_size);
@@ -328,7 +340,7 @@ AllreduceRobust::ReturnType AllreduceRobust::TryResetLinks(void) {
         if (len == sizeof(sig)) all_links[i].size_write = 2;
       }
     }
-    utils::SelectHelper rsel;
+    utils::PollHelper rsel;
     bool finished = true;
     for (int i = 0; i < nlink; ++i) {
       if (all_links[i].size_write != 2 && !all_links[i].sock.BadSocket()) {
@@ -337,15 +349,15 @@ AllreduceRobust::ReturnType AllreduceRobust::TryResetLinks(void) {
     }
     if (finished) break;
     // wait to read from the channels to discard data
-    rsel.Select();
+    rsel.Poll();
   }
   for (int i = 0; i < nlink; ++i) {
     if (!all_links[i].sock.BadSocket()) {
-      utils::SelectHelper::WaitExcept(all_links[i].sock);
+      utils::PollHelper::WaitExcept(all_links[i].sock);
     }
   }
   while (true) {
-    utils::SelectHelper rsel;
+    utils::PollHelper rsel;
     bool finished = true;
     for (int i = 0; i < nlink; ++i) {
       if (all_links[i].size_read == 0 && !all_links[i].sock.BadSocket()) {
@@ -353,7 +365,7 @@ AllreduceRobust::ReturnType AllreduceRobust::TryResetLinks(void) {
       }
     }
     if (finished) break;
-    rsel.Select();
+    rsel.Poll();
     for (int i = 0; i < nlink; ++i) {
       if (all_links[i].sock.BadSocket()) continue;
       if (all_links[i].size_read == 0) {
@@ -618,32 +630,32 @@ AllreduceRobust::TryRecoverData(RecoverType role,
   }
   while (true) {
     bool finished = true;
-    utils::SelectHelper selecter;
+    utils::PollHelper watcher;
     for (int i = 0; i < nlink; ++i) {
       if (i == recv_link && links[i].size_read != size) {
-        selecter.WatchRead(links[i].sock);
+        watcher.WatchRead(links[i].sock);
         finished = false;
       }
       if (req_in[i] && links[i].size_write != size) {
         if (role == kHaveData ||
             (links[recv_link].size_read != links[i].size_write)) {
-          selecter.WatchWrite(links[i].sock);
+          watcher.WatchWrite(links[i].sock);
         }
         finished = false;
       }
-      selecter.WatchException(links[i].sock);
+      watcher.WatchException(links[i].sock);
     }
     if (finished) break;
-    selecter.Select();
+    watcher.Poll();
     // exception handling
     for (int i = 0; i < nlink; ++i) {
-      if (selecter.CheckExcept(links[i].sock)) {
+      if (watcher.CheckExcept(links[i].sock)) {
         return ReportError(&links[i], kGetExcept);
       }
     }
     if (role == kRequestData) {
       const int pid = recv_link;
-      if (selecter.CheckRead(links[pid].sock)) {
+      if (watcher.CheckRead(links[pid].sock)) {
         ReturnType ret = links[pid].ReadToArray(sendrecvbuf_, size);
         if (ret != kSuccess) {
           return ReportError(&links[pid], ret);
@@ -671,7 +683,7 @@ AllreduceRobust::TryRecoverData(RecoverType role,
     if (role == kPassData) {
       const int pid = recv_link;
       const size_t buffer_size = links[pid].buffer_size;
-      if (selecter.CheckRead(links[pid].sock)) {
+      if (watcher.CheckRead(links[pid].sock)) {
         size_t min_write = size;
         for (int i = 0; i < nlink; ++i) {
           if (req_in[i]) min_write = std::min(links[i].size_write, min_write);
@@ -1138,22 +1150,22 @@ AllreduceRobust::RingPassing(void *sendrecvbuf_,
   char *buf = reinterpret_cast<char*>(sendrecvbuf_);
   while (true) {
     bool finished = true;
-    utils::SelectHelper selecter;
+    utils::PollHelper watcher;
     if (read_ptr != read_end) {
-      selecter.WatchRead(prev.sock);
+      watcher.WatchRead(prev.sock);
       finished = false;
     }
     if (write_ptr < read_ptr && write_ptr != write_end) {
-      selecter.WatchWrite(next.sock);
+      watcher.WatchWrite(next.sock);
       finished = false;
     }
-    selecter.WatchException(prev.sock);
-    selecter.WatchException(next.sock);
+    watcher.WatchException(prev.sock);
+    watcher.WatchException(next.sock);
     if (finished) break;
-    selecter.Select();
-    if (selecter.CheckExcept(prev.sock)) return ReportError(&prev, kGetExcept);
-    if (selecter.CheckExcept(next.sock)) return ReportError(&next, kGetExcept);
-    if (read_ptr != read_end && selecter.CheckRead(prev.sock)) {
+    watcher.Poll();
+    if (watcher.CheckExcept(prev.sock)) return ReportError(&prev, kGetExcept);
+    if (watcher.CheckExcept(next.sock)) return ReportError(&next, kGetExcept);
+    if (read_ptr != read_end && watcher.CheckRead(prev.sock)) {
       ssize_t len = prev.sock.Recv(buf + read_ptr, read_end - read_ptr);
       if (len == 0) {
         prev.sock.Close(); return ReportError(&prev, kRecvZeroLen);
