@@ -4,20 +4,29 @@
  * \brief Utilities to store histograms
  * \author Philip Cho, Tianqi Chen
  */
+#include <rabit/rabit.h>
 #include <dmlc/omp.h>
 #include <numeric>
 #include <vector>
-#include "./sync.h"
+
 #include "./random.h"
 #include "./column_matrix.h"
 #include "./hist_util.h"
 #include "./quantile.h"
 
+#if defined(XGBOOST_MM_PREFETCH_PRESENT)
+  #include <xmmintrin.h>
+  #define PREFETCH_READ_T0(addr) _mm_prefetch(reinterpret_cast<const char*>(addr), _MM_HINT_T0)
+#elif defined(XGBOOST_BUILTIN_PREFETCH_PRESENT)
+  #define PREFETCH_READ_T0(addr) __builtin_prefetch(reinterpret_cast<const char*>(addr), 0, 3)
+#else  // no SW pre-fetching available; PREFETCH_READ_T0 is no-op
+  #define PREFETCH_READ_T0(addr) do {} while (0)
+#endif
+
 namespace xgboost {
 namespace common {
 
 void HistCutMatrix::Init(DMatrix* p_fmat, uint32_t max_num_bins) {
-  using WXQSketch = common::WXQuantileSketch<bst_float, bst_float>;
   const MetaInfo& info = p_fmat->Info();
 
   // safe factor for better accuracy
@@ -33,28 +42,37 @@ void HistCutMatrix::Init(DMatrix* p_fmat, uint32_t max_num_bins) {
     s.Init(info.num_row_, 1.0 / (max_num_bins * kFactor));
   }
 
-  auto iter = p_fmat->RowIterator();
-  iter->BeforeFirst();
-  while (iter->Next()) {
-     auto batch = iter->Value();
+  const auto& weights = info.weights_.HostVector();
+  for (const auto &batch : p_fmat->GetRowBatches()) {
     #pragma omp parallel num_threads(nthread)
     {
       CHECK_EQ(nthread, omp_get_num_threads());
       auto tid = static_cast<unsigned>(omp_get_thread_num());
       unsigned begin = std::min(nstep * tid, ncol);
       unsigned end = std::min(nstep * (tid + 1), ncol);
-      for (size_t i = 0; i < batch.Size(); ++i) { // NOLINT(*)
-        size_t ridx = batch.base_rowid + i;
-        SparsePage::Inst inst = batch[i];
-        for (bst_uint j = 0; j < inst.length; ++j) {
-          if (inst[j].index >= begin && inst[j].index < end) {
-            sketchs[inst[j].index].Push(inst[j].fvalue, info.GetWeight(ridx));
+      // do not iterate if no columns are assigned to the thread
+      if (begin < end && end <= ncol) {
+        for (size_t i = 0; i < batch.Size(); ++i) { // NOLINT(*)
+          size_t ridx = batch.base_rowid + i;
+          SparsePage::Inst inst = batch[i];
+          for (auto& ins : inst) {
+            if (ins.index >= begin && ins.index < end) {
+              sketchs[ins.index].Push(ins.fvalue,
+                                      weights.size() > 0 ? weights[ridx] : 1.0f);
+            }
           }
         }
       }
     }
   }
 
+  Init(&sketchs, max_num_bins);
+}
+
+void HistCutMatrix::Init
+(std::vector<WXQSketch>* in_sketchs, uint32_t max_num_bins) {
+  std::vector<WXQSketch>& sketchs = *in_sketchs;
+  constexpr int kFactor = 8;
   // gather the histogram data
   rabit::SerializeReducer<WXQSketch::SummaryContainer> sreducer;
   std::vector<WXQSketch::SummaryContainer> summary_array;
@@ -68,7 +86,7 @@ void HistCutMatrix::Init(DMatrix* p_fmat, uint32_t max_num_bins) {
   size_t nbytes = WXQSketch::SummaryContainer::CalcMemCost(max_num_bins * kFactor);
   sreducer.Allreduce(dmlc::BeginPtr(summary_array), nbytes, summary_array.size());
 
-  this->min_val.resize(info.num_col_);
+  this->min_val.resize(sketchs.size());
   row_ptr.push_back(0);
   for (size_t fid = 0; fid < summary_array.size(); ++fid) {
     WXQSketch::SummaryContainer a;
@@ -104,27 +122,35 @@ void HistCutMatrix::Init(DMatrix* p_fmat, uint32_t max_num_bins) {
   }
 }
 
-void GHistIndexMatrix::Init(DMatrix* p_fmat) {
-  CHECK(cut != nullptr);  // NOLINT
-  auto iter = p_fmat->RowIterator();
+uint32_t HistCutMatrix::GetBinIdx(const Entry& e) {
+  unsigned fid = e.index;
+  auto cbegin = cut.begin() + row_ptr[fid];
+  auto cend = cut.begin() + row_ptr[fid + 1];
+  CHECK(cbegin != cend);
+  auto it = std::upper_bound(cbegin, cend, e.fvalue);
+  if (it == cend) it = cend - 1;
+  uint32_t idx = static_cast<uint32_t>(it - cut.begin());
+  return idx;
+}
+
+void GHistIndexMatrix::Init(DMatrix* p_fmat, int max_num_bins) {
+  cut.Init(p_fmat, max_num_bins);
 
   const int nthread = omp_get_max_threads();
-  const uint32_t nbins = cut->row_ptr.back();
+  const uint32_t nbins = cut.row_ptr.back();
   hit_count.resize(nbins, 0);
   hit_count_tloc_.resize(nthread * nbins, 0);
 
-  iter->BeforeFirst();
   row_ptr.push_back(0);
-  while (iter->Next()) {
-     auto batch = iter->Value();
+  for (const auto &batch : p_fmat->GetRowBatches()) {
     const size_t rbegin = row_ptr.size() - 1;
     for (size_t i = 0; i < batch.Size(); ++i) {
-      row_ptr.push_back(batch[i].length + row_ptr.back());
+      row_ptr.push_back(batch[i].size() + row_ptr.back());
     }
     index.resize(row_ptr.back());
 
-    CHECK_GT(cut->cut.size(), 0U);
-    CHECK_EQ(cut->row_ptr.back(), cut->cut.size());
+    CHECK_GT(cut.cut.size(), 0U);
+    CHECK_EQ(cut.row_ptr.back(), cut.cut.size());
 
     auto bsize = static_cast<omp_ulong>(batch.Size());
     #pragma omp parallel for num_threads(nthread) schedule(static)
@@ -133,15 +159,11 @@ void GHistIndexMatrix::Init(DMatrix* p_fmat) {
       size_t ibegin = row_ptr[rbegin + i];
       size_t iend = row_ptr[rbegin + i + 1];
       SparsePage::Inst inst = batch[i];
-      CHECK_EQ(ibegin + inst.length, iend);
-      for (bst_uint j = 0; j < inst.length; ++j) {
-        unsigned fid = inst[j].index;
-        auto cbegin = cut->cut.begin() + cut->row_ptr[fid];
-        auto cend = cut->cut.begin() + cut->row_ptr[fid + 1];
-        CHECK(cbegin != cend);
-        auto it = std::upper_bound(cbegin, cend, inst[j].fvalue);
-        if (it == cend) it = cend - 1;
-        uint32_t idx = static_cast<uint32_t>(it - cut->cut.begin());
+
+      CHECK_EQ(ibegin + inst.size(), iend);
+      for (bst_uint j = 0; j < inst.size(); ++j) {
+        uint32_t idx = cut.GetBinIdx(inst[j]);
+
         index[ibegin + j] = idx;
         ++hit_count_tloc_[tid * nbins + idx];
       }
@@ -157,14 +179,13 @@ void GHistIndexMatrix::Init(DMatrix* p_fmat) {
   }
 }
 
-template <typename T>
 static size_t GetConflictCount(const std::vector<bool>& mark,
-                               const Column<T>& column,
+                               const Column& column,
                                size_t max_cnt) {
   size_t ret = 0;
-  if (column.type == xgboost::common::kDenseColumn) {
-    for (size_t i = 0; i < column.len; ++i) {
-      if (column.index[i] != std::numeric_limits<T>::max() && mark[i]) {
+  if (column.GetType() == xgboost::common::kDenseColumn) {
+    for (size_t i = 0; i < column.Size(); ++i) {
+      if (column.GetFeatureBinIdx(i) != std::numeric_limits<uint32_t>::max() && mark[i]) {
         ++ret;
         if (ret > max_cnt) {
           return max_cnt + 1;
@@ -172,8 +193,8 @@ static size_t GetConflictCount(const std::vector<bool>& mark,
       }
     }
   } else {
-    for (size_t i = 0; i < column.len; ++i) {
-      if (mark[column.row_ind[i]]) {
+    for (size_t i = 0; i < column.Size(); ++i) {
+      if (mark[column.GetRowIdx(i)]) {
         ++ret;
         if (ret > max_cnt) {
           return max_cnt + 1;
@@ -184,30 +205,28 @@ static size_t GetConflictCount(const std::vector<bool>& mark,
   return ret;
 }
 
-template <typename T>
 inline void
-MarkUsed(std::vector<bool>* p_mark, const Column<T>& column) {
+MarkUsed(std::vector<bool>* p_mark, const Column& column) {
   std::vector<bool>& mark = *p_mark;
-  if (column.type == xgboost::common::kDenseColumn) {
-    for (size_t i = 0; i < column.len; ++i) {
-      if (column.index[i] != std::numeric_limits<T>::max()) {
+  if (column.GetType() == xgboost::common::kDenseColumn) {
+    for (size_t i = 0; i < column.Size(); ++i) {
+      if (column.GetFeatureBinIdx(i) != std::numeric_limits<uint32_t>::max()) {
         mark[i] = true;
       }
     }
   } else {
-    for (size_t i = 0; i < column.len; ++i) {
-      mark[column.row_ind[i]] = true;
+    for (size_t i = 0; i < column.Size(); ++i) {
+      mark[column.GetRowIdx(i)] = true;
     }
   }
 }
 
-template <typename T>
 inline std::vector<std::vector<unsigned>>
-FindGroups_(const std::vector<unsigned>& feature_list,
-            const std::vector<size_t>& feature_nnz,
-            const ColumnMatrix& colmat,
-            size_t nrow,
-            const FastHistParam& param) {
+FindGroups(const std::vector<unsigned>& feature_list,
+           const std::vector<size_t>& feature_nnz,
+           const ColumnMatrix& colmat,
+           size_t nrow,
+           const tree::TrainParam& param) {
   /* Goal: Bundle features together that has little or no "overlap", i.e.
            only a few data points should have nonzero values for
            member features.
@@ -221,7 +240,7 @@ FindGroups_(const std::vector<unsigned>& feature_list,
     = static_cast<size_t>(param.max_conflict_rate * nrow);
 
   for (auto fid : feature_list) {
-    const Column<T>& column = colmat.GetColumn<T>(fid);
+    const Column& column = colmat.GetColumn(fid);
 
     const size_t cur_fid_nnz = feature_nnz[fid];
     bool need_new_group = true;
@@ -266,24 +285,12 @@ FindGroups_(const std::vector<unsigned>& feature_list,
   return groups;
 }
 
-inline std::vector<std::vector<unsigned>>
-FindGroups(const std::vector<unsigned>& feature_list,
-           const std::vector<size_t>& feature_nnz,
-           const ColumnMatrix& colmat,
-           size_t nrow,
-           const FastHistParam& param) {
-  XGBOOST_TYPE_SWITCH(colmat.dtype, {
-    return FindGroups_<DType>(feature_list, feature_nnz, colmat, nrow, param);
-  });
-  return std::vector<std::vector<unsigned>>();  // to avoid warning message
-}
-
 inline std::vector<std::vector<unsigned>>
 FastFeatureGrouping(const GHistIndexMatrix& gmat,
                     const ColumnMatrix& colmat,
-                    const FastHistParam& param) {
+                    const tree::TrainParam& param) {
   const size_t nrow = gmat.row_ptr.size() - 1;
-  const size_t nfeature = gmat.cut->row_ptr.size() - 1;
+  const size_t nfeature = gmat.cut.row_ptr.size() - 1;
 
   std::vector<unsigned> feature_list(nfeature);
   std::iota(feature_list.begin(), feature_list.end(), 0);
@@ -335,11 +342,11 @@ FastFeatureGrouping(const GHistIndexMatrix& gmat,
 
 void GHistIndexBlockMatrix::Init(const GHistIndexMatrix& gmat,
                                  const ColumnMatrix& colmat,
-                                 const FastHistParam& param) {
-  cut_ = gmat.cut;
+                                 const tree::TrainParam& param) {
+  cut_ = &gmat.cut;
 
   const size_t nrow = gmat.row_ptr.size() - 1;
-  const uint32_t nbins = gmat.cut->row_ptr.back();
+  const uint32_t nbins = gmat.cut.row_ptr.back();
 
   /* step 1: form feature groups */
   auto groups = FastFeatureGrouping(gmat, colmat, param);
@@ -349,8 +356,8 @@ void GHistIndexBlockMatrix::Init(const GHistIndexMatrix& gmat,
   std::vector<uint32_t> bin2block(nbins);  // lookup table [bin id] => [block id]
   for (uint32_t group_id = 0; group_id < nblock; ++group_id) {
     for (auto& fid : groups[group_id]) {
-      const uint32_t bin_begin = gmat.cut->row_ptr[fid];
-      const uint32_t bin_end = gmat.cut->row_ptr[fid + 1];
+      const uint32_t bin_begin = gmat.cut.row_ptr[fid];
+      const uint32_t bin_end = gmat.cut.row_ptr[fid + 1];
       for (uint32_t bin_id = bin_begin; bin_id < bin_end; ++bin_id) {
         bin2block[bin_id] = group_id;
       }
@@ -400,58 +407,90 @@ void GHistIndexBlockMatrix::Init(const GHistIndexMatrix& gmat,
 void GHistBuilder::BuildHist(const std::vector<GradientPair>& gpair,
                              const RowSetCollection::Elem row_indices,
                              const GHistIndexMatrix& gmat,
-                             const std::vector<bst_uint>& feat_set,
                              GHistRow hist) {
-  data_.resize(nbins_ * nthread_, GHistEntry());
-  std::fill(data_.begin(), data_.end(), GHistEntry());
+  const size_t nthread = static_cast<size_t>(this->nthread_);
+  data_.resize(nbins_ * nthread_);
 
-  constexpr int kUnroll = 8;  // loop unrolling factor
-  const auto nthread = static_cast<bst_omp_uint>(this->nthread_);
-  const size_t nrows = row_indices.end - row_indices.begin;
-  const size_t rest = nrows % kUnroll;
+  const size_t* rid =  row_indices.begin;
+  const size_t nrows = row_indices.Size();
+  const uint32_t* index = gmat.index.data();
+  const size_t* row_ptr =  gmat.row_ptr.data();
+  const float* pgh = reinterpret_cast<const float*>(gpair.data());
 
-  #pragma omp parallel for num_threads(nthread) schedule(guided)
-  for (bst_omp_uint i = 0; i < nrows - rest; i += kUnroll) {
-    const bst_omp_uint tid = omp_get_thread_num();
-    const size_t off = tid * nbins_;
-    size_t rid[kUnroll];
-    size_t ibegin[kUnroll];
-    size_t iend[kUnroll];
-    GradientPair stat[kUnroll];
-    for (int k = 0; k < kUnroll; ++k) {
-      rid[k] = row_indices.begin[i + k];
+  double* hist_data = reinterpret_cast<double*>(hist.begin);
+  double* data = reinterpret_cast<double*>(data_.data());
+
+  const size_t block_size = 512;
+  size_t n_blocks = nrows/block_size;
+  n_blocks += !!(nrows - n_blocks*block_size);
+
+  const size_t nthread_to_process = std::min(nthread,  n_blocks);
+  memset(thread_init_.data(), '\0', nthread_to_process*sizeof(size_t));
+
+  const size_t cache_line_size = 64;
+  const size_t prefetch_offset = 10;
+  size_t no_prefetch_size = prefetch_offset + cache_line_size/sizeof(*rid);
+  no_prefetch_size = no_prefetch_size > nrows ? nrows : no_prefetch_size;
+
+  #pragma omp parallel for num_threads(nthread_to_process) schedule(guided)
+  for (bst_omp_uint iblock = 0; iblock < n_blocks; iblock++) {
+    dmlc::omp_uint tid = omp_get_thread_num();
+    double* data_local_hist = ((nthread_to_process == 1) ? hist_data :
+            reinterpret_cast<double*>(data_.data() + tid * nbins_));
+
+    if (!thread_init_[tid]) {
+      memset(data_local_hist, '\0', 2*nbins_*sizeof(double));
+      thread_init_[tid] = true;
     }
-    for (int k = 0; k < kUnroll; ++k) {
-      ibegin[k] = gmat.row_ptr[rid[k]];
-      iend[k] = gmat.row_ptr[rid[k] + 1];
-    }
-    for (int k = 0; k < kUnroll; ++k) {
-      stat[k] = gpair[rid[k]];
-    }
-    for (int k = 0; k < kUnroll; ++k) {
-      for (size_t j = ibegin[k]; j < iend[k]; ++j) {
-        const uint32_t bin = gmat.index[j];
-        data_[off + bin].Add(stat[k]);
+
+    const size_t istart = iblock*block_size;
+    const size_t iend = (((iblock+1)*block_size > nrows) ? nrows : istart + block_size);
+    for (size_t i = istart; i < iend; ++i) {
+      const size_t icol_start = row_ptr[rid[i]];
+      const size_t icol_end = row_ptr[rid[i]+1];
+
+      if (i < nrows - no_prefetch_size) {
+        PREFETCH_READ_T0(row_ptr + rid[i + prefetch_offset]);
+        PREFETCH_READ_T0(pgh + 2*rid[i + prefetch_offset]);
+      }
+
+      for (size_t j = icol_start; j < icol_end; ++j) {
+        const uint32_t idx_bin = 2*index[j];
+        const size_t idx_gh = 2*rid[i];
+
+        data_local_hist[idx_bin] += pgh[idx_gh];
+        data_local_hist[idx_bin+1] += pgh[idx_gh+1];
       }
     }
   }
-  for (size_t i = nrows - rest; i < nrows; ++i) {
-    const size_t rid = row_indices.begin[i];
-    const size_t ibegin = gmat.row_ptr[rid];
-    const size_t iend = gmat.row_ptr[rid + 1];
-    const GradientPair stat = gpair[rid];
-    for (size_t j = ibegin; j < iend; ++j) {
-      const uint32_t bin = gmat.index[j];
-      data_[bin].Add(stat);
-    }
-  }
 
-  /* reduction */
-  const uint32_t nbins = nbins_;
-  #pragma omp parallel for num_threads(nthread) schedule(static)
-  for (bst_omp_uint bin_id = 0; bin_id < bst_omp_uint(nbins); ++bin_id) {
-    for (bst_omp_uint tid = 0; tid < nthread; ++tid) {
-      hist.begin[bin_id].Add(data_[tid * nbins_ + bin_id]);
+  if (nthread_to_process > 1) {
+    const size_t size = (2*nbins_);
+    const size_t block_size = 1024;
+    size_t n_blocks = size/block_size;
+    n_blocks += !!(size - n_blocks*block_size);
+
+    size_t n_worked_bins = 0;
+    for (size_t i = 0; i < nthread_to_process; ++i) {
+      if (thread_init_[i]) {
+        thread_init_[n_worked_bins++] = i;
+      }
+    }
+
+    #pragma omp parallel for num_threads(std::min(nthread, n_blocks)) schedule(guided)
+    for (bst_omp_uint iblock = 0; iblock < n_blocks; iblock++) {
+      const size_t istart = iblock*block_size;
+      const size_t iend = (((iblock+1)*block_size > size) ? size : istart + block_size);
+
+      const size_t bin = 2*thread_init_[0]*nbins_;
+      memcpy(hist_data + istart, (data + bin + istart), sizeof(double)*(iend - istart));
+
+      for (size_t i_bin_part = 1; i_bin_part < n_worked_bins; ++i_bin_part) {
+        const size_t bin = 2*thread_init_[i_bin_part]*nbins_;
+        for (size_t i = istart; i < iend; i++) {
+          hist_data[i] += data[bin + i];
+        }
+      }
     }
   }
 }
@@ -459,7 +498,6 @@ void GHistBuilder::BuildHist(const std::vector<GradientPair>& gpair,
 void GHistBuilder::BuildBlockHist(const std::vector<GradientPair>& gpair,
                                   const RowSetCollection::Elem row_indices,
                                   const GHistIndexBlockMatrix& gmatb,
-                                  const std::vector<bst_uint>& feat_set,
                                   GHistRow hist) {
   constexpr int kUnroll = 8;  // loop unrolling factor
   const size_t nblock = gmatb.GetNumBlock();
