@@ -30,11 +30,15 @@ ObjFunction* ObjFunction::Create(const std::string& name) {
 namespace xgboost {
 namespace obj {
 // List of files that will be force linked in static links.
-DMLC_REGISTRY_LINK_TAG(regression_obj);
 #ifdef XGBOOST_USE_CUDA
-  DMLC_REGISTRY_LINK_TAG(regression_obj_gpu);
-#endif
+DMLC_REGISTRY_LINK_TAG(regression_obj_gpu);
+DMLC_REGISTRY_LINK_TAG(hinge_obj_gpu);
+DMLC_REGISTRY_LINK_TAG(multiclass_obj_gpu);
+#else
+DMLC_REGISTRY_LINK_TAG(regression_obj);
+DMLC_REGISTRY_LINK_TAG(hinge_obj);
 DMLC_REGISTRY_LINK_TAG(multiclass_obj);
+#endif
 DMLC_REGISTRY_LINK_TAG(rank_obj);
 }  // namespace obj
 }  // namespace xgboost
