@@ -4,15 +4,17 @@
  * \brief use columnwise update to construct a tree
  * \author Tianqi Chen
  */
+#include <rabit/rabit.h>
 #include <xgboost/tree_updater.h>
 #include <memory>
 #include <vector>
 #include <cmath>
 #include <algorithm>
+
 #include "./param.h"
 #include "../common/random.h"
 #include "../common/bitmap.h"
-#include "../common/sync.h"
+#include "split_evaluator.h"
 
 namespace xgboost {
 namespace tree {
@@ -20,25 +22,27 @@ namespace tree {
 DMLC_REGISTRY_FILE_TAG(updater_colmaker);
 
 /*! \brief column-wise update to construct a tree */
-template<typename TStats, typename TConstraint>
 class ColMaker: public TreeUpdater {
  public:
   void Init(const std::vector<std::pair<std::string, std::string> >& args) override {
     param_.InitAllowUnknown(args);
+    spliteval_.reset(SplitEvaluator::Create(param_.split_evaluator));
+    spliteval_->Init(args);
   }
 
   void Update(HostDeviceVector<GradientPair> *gpair,
               DMatrix* dmat,
               const std::vector<RegTree*> &trees) override {
-    TStats::CheckInfo(dmat->Info());
+    GradStats::CheckInfo(dmat->Info());
     // rescale learning rate according to size of trees
     float lr = param_.learning_rate;
     param_.learning_rate = lr / trees.size();
-    TConstraint::Init(&param_, dmat->Info().num_col_);
     // build tree
     for (auto tree : trees) {
-      Builder builder(param_);
-      builder.Update(gpair->HostVector(), dmat, tree);
+      Builder builder(
+        param_,
+        std::unique_ptr<SplitEvaluator>(spliteval_->GetHostClone()));
+      builder.Update(gpair->ConstHostVector(), dmat, tree);
     }
     param_.learning_rate = lr;
   }
@@ -46,13 +50,15 @@ class ColMaker: public TreeUpdater {
  protected:
   // training parameter
   TrainParam param_;
+  // SplitEvaluator that will be cloned for each Builder
+  std::unique_ptr<SplitEvaluator> spliteval_;
   // data structure
   /*! \brief per thread x per node entry to store tmp data */
   struct ThreadEntry {
     /*! \brief statistics of data */
-    TStats stats;
+    GradStats stats;
     /*! \brief extra statistics of data */
-    TStats stats_extra;
+    GradStats stats_extra;
     /*! \brief last feature value scanned */
     bst_float last_fvalue;
     /*! \brief first feature value scanned */
@@ -66,7 +72,7 @@ class ColMaker: public TreeUpdater {
   };
   struct NodeEntry {
     /*! \brief statics for node entry */
-    TStats stats;
+    GradStats stats;
     /*! \brief loss of this node, without split */
     bst_float root_gain;
     /*! \brief weight calculated related to current data */
@@ -82,24 +88,41 @@ class ColMaker: public TreeUpdater {
   class Builder {
    public:
     // constructor
-    explicit Builder(const TrainParam& param) : param_(param), nthread_(omp_get_max_threads()) {}
+    explicit Builder(const TrainParam& param,
+                     std::unique_ptr<SplitEvaluator> spliteval)
+        : param_(param), nthread_(omp_get_max_threads()),
+          spliteval_(std::move(spliteval)) {}
     // update one tree, growing
     virtual void Update(const std::vector<GradientPair>& gpair,
                         DMatrix* p_fmat,
                         RegTree* p_tree) {
+      std::vector<int> newnodes;
       this->InitData(gpair, *p_fmat, *p_tree);
       this->InitNewNode(qexpand_, gpair, *p_fmat, *p_tree);
       for (int depth = 0; depth < param_.max_depth; ++depth) {
         this->FindSplit(depth, qexpand_, gpair, p_fmat, p_tree);
         this->ResetPosition(qexpand_, p_fmat, *p_tree);
-        this->UpdateQueueExpand(*p_tree, &qexpand_);
-        this->InitNewNode(qexpand_, gpair, *p_fmat, *p_tree);
+        this->UpdateQueueExpand(*p_tree, qexpand_, &newnodes);
+        this->InitNewNode(newnodes, gpair, *p_fmat, *p_tree);
+        for (auto nid : qexpand_) {
+          if ((*p_tree)[nid].IsLeaf()) {
+            continue;
+          }
+          int cleft = (*p_tree)[nid].LeftChild();
+          int cright = (*p_tree)[nid].RightChild();
+          spliteval_->AddSplit(nid,
+                               cleft,
+                               cright,
+                               snode_[nid].best.SplitIndex(),
+                               snode_[cleft].weight,
+                               snode_[cright].weight);
+        }
+        qexpand_ = newnodes;
         // if nothing left to be expand, break
         if (qexpand_.size() == 0) break;
       }
       // set all the rest expanding nodes to leaf
-      for (size_t i = 0; i < qexpand_.size(); ++i) {
-        const int nid = qexpand_[i];
+      for (const int nid : qexpand_) {
         (*p_tree)[nid].SetLeaf(snode_[nid].weight * param_.learning_rate);
       }
       // remember auxiliary statistics in the tree node
@@ -107,7 +130,6 @@ class ColMaker: public TreeUpdater {
         p_tree->Stat(nid).loss_chg = snode_[nid].best.loss_chg;
         p_tree->Stat(nid).base_weight = snode_[nid].weight;
         p_tree->Stat(nid).sum_hess = static_cast<float>(snode_[nid].stats.sum_hess);
-        snode_[nid].stats.SetLeafVec(param_, p_tree->Leafvec(nid));
       }
     }
 
@@ -119,59 +141,43 @@ class ColMaker: public TreeUpdater {
       CHECK_EQ(tree.param.num_nodes, tree.param.num_roots)
           << "ColMaker: can only grow new tree";
       const std::vector<unsigned>& root_index = fmat.Info().root_index_;
-      const RowSet& rowset = fmat.BufferedRowset();
       {
         // setup position
         position_.resize(gpair.size());
+        CHECK_EQ(fmat.Info().num_row_, position_.size());
         if (root_index.size() == 0) {
-          for (size_t i = 0; i < rowset.Size(); ++i) {
-            position_[rowset[i]] = 0;
-          }
+          std::fill(position_.begin(), position_.end(), 0);
         } else {
-          for (size_t i = 0; i < rowset.Size(); ++i) {
-            const bst_uint ridx = rowset[i];
+          for (size_t ridx = 0; ridx <  position_.size(); ++ridx) {
             position_[ridx] = root_index[ridx];
             CHECK_LT(root_index[ridx], (unsigned)tree.param.num_roots);
           }
         }
         // mark delete for the deleted datas
-        for (size_t i = 0; i < rowset.Size(); ++i) {
-          const bst_uint ridx = rowset[i];
+        for (size_t ridx = 0; ridx < position_.size(); ++ridx) {
           if (gpair[ridx].GetHess() < 0.0f) position_[ridx] = ~position_[ridx];
         }
         // mark subsample
         if (param_.subsample < 1.0f) {
           std::bernoulli_distribution coin_flip(param_.subsample);
           auto& rnd = common::GlobalRandom();
-          for (size_t i = 0; i < rowset.Size(); ++i) {
-            const bst_uint ridx = rowset[i];
+          for (size_t ridx = 0; ridx < position_.size(); ++ridx) {
             if (gpair[ridx].GetHess() < 0.0f) continue;
             if (!coin_flip(rnd)) position_[ridx] = ~position_[ridx];
           }
         }
       }
       {
-        // initialize feature index
-        auto ncol = static_cast<unsigned>(fmat.Info().num_col_);
-        for (unsigned i = 0; i < ncol; ++i) {
-          if (fmat.GetColSize(i) != 0) {
-            feat_index_.push_back(i);
-          }
-        }
-        unsigned n = std::max(static_cast<unsigned>(1),
-                              static_cast<unsigned>(param_.colsample_bytree * feat_index_.size()));
-        std::shuffle(feat_index_.begin(), feat_index_.end(), common::GlobalRandom());
-        CHECK_GT(param_.colsample_bytree, 0U)
-            << "colsample_bytree cannot be zero.";
-        feat_index_.resize(n);
+        column_sampler_.Init(fmat.Info().num_col_, param_.colsample_bynode,
+                             param_.colsample_bylevel, param_.colsample_bytree);
       }
       {
         // setup temp space for each thread
         // reserve a small space
         stemp_.clear();
         stemp_.resize(this->nthread_, std::vector<ThreadEntry>());
-        for (size_t i = 0; i < stemp_.size(); ++i) {
-          stemp_[i].clear(); stemp_[i].reserve(256);
+        for (auto& i : stemp_) {
+          i.clear(); i.reserve(256);
         }
         snode_.reserve(256);
       }
@@ -193,74 +199,61 @@ class ColMaker: public TreeUpdater {
                             const RegTree& tree) {
       {
         // setup statistics space for each tree node
-        for (size_t i = 0; i < stemp_.size(); ++i) {
-          stemp_[i].resize(tree.param.num_nodes, ThreadEntry(param_));
+        for (auto& i : stemp_) {
+          i.resize(tree.param.num_nodes, ThreadEntry(param_));
         }
         snode_.resize(tree.param.num_nodes, NodeEntry(param_));
-        constraints_.resize(tree.param.num_nodes);
       }
-      const RowSet &rowset = fmat.BufferedRowset();
       const MetaInfo& info = fmat.Info();
       // setup position
-      const auto ndata = static_cast<bst_omp_uint>(rowset.Size());
+      const auto ndata = static_cast<bst_omp_uint>(info.num_row_);
       #pragma omp parallel for schedule(static)
-      for (bst_omp_uint i = 0; i < ndata; ++i) {
-        const bst_uint ridx = rowset[i];
+      for (bst_omp_uint ridx = 0; ridx < ndata; ++ridx) {
         const int tid = omp_get_thread_num();
         if (position_[ridx] < 0) continue;
         stemp_[tid][position_[ridx]].stats.Add(gpair, info, ridx);
       }
       // sum the per thread statistics together
       for (int nid : qexpand) {
-        TStats stats(param_);
-        for (size_t tid = 0; tid < stemp_.size(); ++tid) {
-          stats.Add(stemp_[tid][nid].stats);
+        GradStats stats(param_);
+        for (auto& s : stemp_) {
+          stats.Add(s[nid].stats);
         }
         // update node statistics
         snode_[nid].stats = stats;
       }
-      // setup constraints before calculating the weight
-      for (int nid : qexpand) {
-        if (tree[nid].IsRoot()) continue;
-        const int pid = tree[nid].Parent();
-        constraints_[pid].SetChild(param_, tree[pid].SplitIndex(),
-                                   snode_[tree[pid].LeftChild()].stats,
-                                   snode_[tree[pid].RightChild()].stats,
-                                   &constraints_[tree[pid].LeftChild()],
-                                   &constraints_[tree[pid].RightChild()]);
-      }
       // calculating the weights
       for (int nid : qexpand) {
-        snode_[nid].root_gain = static_cast<float>(
-            constraints_[nid].CalcGain(param_, snode_[nid].stats));
+        bst_uint parentid = tree[nid].Parent();
         snode_[nid].weight = static_cast<float>(
-            constraints_[nid].CalcWeight(param_, snode_[nid].stats));
+            spliteval_->ComputeWeight(parentid, snode_[nid].stats));
+        snode_[nid].root_gain = static_cast<float>(
+            spliteval_->ComputeScore(parentid, snode_[nid].stats, snode_[nid].weight));
       }
     }
     /*! \brief update queue expand add in new leaves */
-    inline void UpdateQueueExpand(const RegTree& tree, std::vector<int>* p_qexpand) {
-      std::vector<int> &qexpand = *p_qexpand;
-      std::vector<int> newnodes;
+    inline void UpdateQueueExpand(const RegTree& tree,
+                                  const std::vector<int> &qexpand,
+                                  std::vector<int>* p_newnodes) {
+      p_newnodes->clear();
       for (int nid : qexpand) {
         if (!tree[ nid ].IsLeaf()) {
-          newnodes.push_back(tree[nid].LeftChild());
-          newnodes.push_back(tree[nid].RightChild());
+          p_newnodes->push_back(tree[nid].LeftChild());
+          p_newnodes->push_back(tree[nid].RightChild());
         }
       }
-      // use new nodes for qexpand
-      qexpand = newnodes;
     }
     // parallel find the best split of current fid
     // this function does not support nested functions
     inline void ParallelFindSplit(const SparsePage::Inst &col,
                                   bst_uint fid,
-                                  const DMatrix &fmat,
+                                 DMatrix *p_fmat,
                                   const std::vector<GradientPair> &gpair) {
       // TODO(tqchen): double check stats order.
-      const MetaInfo& info = fmat.Info();
-      const bool ind = col.length != 0 && col.data[0].fvalue == col.data[col.length - 1].fvalue;
-      bool need_forward = param_.NeedForwardSearch(fmat.GetColDensity(fid), ind);
-      bool need_backward = param_.NeedBackwardSearch(fmat.GetColDensity(fid), ind);
+      const MetaInfo& info = p_fmat->Info();
+      const bool ind = col.size() != 0 && col[0].fvalue == col[col.size() - 1].fvalue;
+      bool need_forward = param_.NeedForwardSearch(p_fmat->GetColDensity(fid), ind);
+      bool need_backward = param_.NeedBackwardSearch(p_fmat->GetColDensity(fid), ind);
       const std::vector<int> &qexpand = qexpand_;
       #pragma omp parallel
       {
@@ -270,8 +263,8 @@ class ColMaker: public TreeUpdater {
         for (int j : qexpand) {
           temp[j].stats.Clear();
         }
-        bst_uint step = (col.length + this->nthread_ - 1) / this->nthread_;
-        bst_uint end = std::min(col.length, step * (tid + 1));
+        bst_uint step = (col.size() + this->nthread_ - 1) / this->nthread_;
+        bst_uint end = std::min(static_cast<bst_uint>(col.size()), step * (tid + 1));
         for (bst_uint i = tid * step; i < end; ++i) {
           const bst_uint ridx = col[i].index;
           const int nid = position_[ridx];
@@ -289,7 +282,7 @@ class ColMaker: public TreeUpdater {
       #pragma omp parallel for schedule(static)
       for (bst_omp_uint j = 0; j < nnode; ++j) {
         const int nid = qexpand[j];
-        TStats sum(param_), tmp(param_), c(param_);
+        GradStats sum(param_), tmp(param_), c(param_);
         for (int tid = 0; tid < this->nthread_; ++tid) {
           tmp = stemp_[tid][nid].stats;
           stemp_[tid][nid].stats = sum;
@@ -316,10 +309,9 @@ class ColMaker: public TreeUpdater {
             if (c.sum_hess >= param_.min_child_weight &&
                 e.stats.sum_hess >= param_.min_child_weight) {
               auto loss_chg = static_cast<bst_float>(
-                  constraints_[nid].CalcSplitGain(
-                      param_, param_.monotone_constraints[fid], e.stats, c) -
+                  spliteval_->ComputeSplitScore(nid, fid, e.stats, c) -
                   snode_[nid].root_gain);
-              e.best.Update(loss_chg, fid, fsplit, false);
+              e.best.Update(loss_chg, fid, fsplit, false, e.stats, c);
             }
           }
           if (need_backward) {
@@ -328,10 +320,9 @@ class ColMaker: public TreeUpdater {
             if (c.sum_hess >= param_.min_child_weight &&
                 tmp.sum_hess >= param_.min_child_weight) {
               auto loss_chg = static_cast<bst_float>(
-                  constraints_[nid].CalcSplitGain(
-                      param_, param_.monotone_constraints[fid], tmp, c) -
+                  spliteval_->ComputeSplitScore(nid, fid, tmp, c) -
                   snode_[nid].root_gain);
-              e.best.Update(loss_chg, fid, fsplit, true);
+              e.best.Update(loss_chg, fid, fsplit, true, tmp, c);
             }
           }
         }
@@ -342,21 +333,20 @@ class ColMaker: public TreeUpdater {
           if (c.sum_hess >= param_.min_child_weight &&
               tmp.sum_hess >= param_.min_child_weight) {
             auto loss_chg = static_cast<bst_float>(
-                constraints_[nid].CalcSplitGain(
-                    param_, param_.monotone_constraints[fid], tmp, c) -
+                spliteval_->ComputeSplitScore(nid, fid, tmp, c) -
                 snode_[nid].root_gain);
-            e.best.Update(loss_chg, fid, e.last_fvalue + kRtEps, true);
+            e.best.Update(loss_chg, fid, e.last_fvalue + kRtEps, true, tmp, c);
           }
         }
       }
       // rescan, generate candidate split
       #pragma omp parallel
       {
-        TStats c(param_), cright(param_);
+        GradStats c(param_), cright(param_);
         const int tid = omp_get_thread_num();
         std::vector<ThreadEntry> &temp = stemp_[tid];
-        bst_uint step = (col.length + this->nthread_ - 1) / this->nthread_;
-        bst_uint end = std::min(col.length, step * (tid + 1));
+        bst_uint step = (col.size() + this->nthread_ - 1) / this->nthread_;
+        bst_uint end = std::min(static_cast<bst_uint>(col.size()), step * (tid + 1));
         for (bst_uint i = tid * step; i < end; ++i) {
           const bst_uint ridx = col[i].index;
           const int nid = position_[ridx];
@@ -375,11 +365,10 @@ class ColMaker: public TreeUpdater {
                 if (c.sum_hess >= param_.min_child_weight &&
                     e.stats.sum_hess >= param_.min_child_weight) {
                   auto loss_chg = static_cast<bst_float>(
-                      constraints_[nid].CalcSplitGain(
-                          param_, param_.monotone_constraints[fid], e.stats, c) -
+                      spliteval_->ComputeSplitScore(nid, fid, e.stats, c) -
                       snode_[nid].root_gain);
                   e.best.Update(loss_chg, fid, (fvalue + e.first_fvalue) * 0.5f,
-                                false);
+                                false, e.stats, c);
                 }
               }
               if (need_backward) {
@@ -388,10 +377,9 @@ class ColMaker: public TreeUpdater {
                 if (c.sum_hess >= param_.min_child_weight &&
                     cright.sum_hess >= param_.min_child_weight) {
                   auto loss_chg = static_cast<bst_float>(
-                      constraints_[nid].CalcSplitGain(
-                          param_, param_.monotone_constraints[fid], c, cright) -
+                      spliteval_->ComputeSplitScore(nid, fid, c, cright) -
                       snode_[nid].root_gain);
-                  e.best.Update(loss_chg, fid, (fvalue + e.first_fvalue) * 0.5f, true);
+                  e.best.Update(loss_chg, fid, (fvalue + e.first_fvalue) * 0.5f, true, c, cright);
                 }
               }
             }
@@ -404,7 +392,7 @@ class ColMaker: public TreeUpdater {
     // update enumeration solution
     inline void UpdateEnumeration(int nid, GradientPair gstats,
                                   bst_float fvalue, int d_step, bst_uint fid,
-                                  TStats &c, std::vector<ThreadEntry> &temp) { // NOLINT(*)
+                                  GradStats &c, std::vector<ThreadEntry> &temp) { // NOLINT(*)
       // get the statistics of nid
       ThreadEntry &e = temp[nid];
       // test if first hit, this is fine, because we set 0 during init
@@ -420,17 +408,17 @@ class ColMaker: public TreeUpdater {
             bst_float loss_chg;
             if (d_step == -1) {
               loss_chg = static_cast<bst_float>(
-                  constraints_[nid].CalcSplitGain(
-                      param_, param_.monotone_constraints[fid], c, e.stats) -
+                  spliteval_->ComputeSplitScore(nid, fid, c, e.stats) -
                   snode_[nid].root_gain);
+              e.best.Update(loss_chg, fid, (fvalue + e.last_fvalue) * 0.5f,
+                            d_step == -1, c, e.stats);
             } else {
               loss_chg = static_cast<bst_float>(
-                  constraints_[nid].CalcSplitGain(
-                      param_, param_.monotone_constraints[fid], e.stats, c) -
+                  spliteval_->ComputeSplitScore(nid, fid, e.stats, c) -
                   snode_[nid].root_gain);
+              e.best.Update(loss_chg, fid, (fvalue + e.last_fvalue) * 0.5f,
+                            d_step == -1, e.stats, c);
             }
-            e.best.Update(loss_chg, fid, (fvalue + e.last_fvalue) * 0.5f,
-                          d_step == -1);
           }
         }
         // update the statistics
@@ -451,7 +439,7 @@ class ColMaker: public TreeUpdater {
         temp[nid].stats.Clear();
       }
       // left statistics
-      TStats c(param_);
+      GradStats c(param_);
       // local cache buffer for position and gradient pair
       constexpr int kBuffer = 32;
       int buf_position[kBuffer] = {};
@@ -500,20 +488,21 @@ class ColMaker: public TreeUpdater {
         if (e.stats.sum_hess >= param_.min_child_weight &&
             c.sum_hess >= param_.min_child_weight) {
           bst_float loss_chg;
-          if (d_step == -1) {
-            loss_chg = static_cast<bst_float>(
-                constraints_[nid].CalcSplitGain(
-                    param_, param_.monotone_constraints[fid], c, e.stats) -
-                snode_[nid].root_gain);
-          } else {
-            loss_chg = static_cast<bst_float>(
-                constraints_[nid].CalcSplitGain(
-                    param_, param_.monotone_constraints[fid], e.stats, c) -
-                snode_[nid].root_gain);
-          }
           const bst_float gap = std::abs(e.last_fvalue) + kRtEps;
           const bst_float delta = d_step == +1 ? gap: -gap;
-          e.best.Update(loss_chg, fid, e.last_fvalue + delta, d_step == -1);
+          if (d_step == -1) {
+            loss_chg = static_cast<bst_float>(
+                spliteval_->ComputeSplitScore(nid, fid, c, e.stats) -
+                snode_[nid].root_gain);
+            e.best.Update(loss_chg, fid, e.last_fvalue + delta, d_step == -1, c,
+                          e.stats);
+          } else {
+            loss_chg = static_cast<bst_float>(
+                spliteval_->ComputeSplitScore(nid, fid, e.stats, c) -
+                snode_[nid].root_gain);
+            e.best.Update(loss_chg, fid, e.last_fvalue + delta, d_step == -1,
+                          e.stats, c);
+          }
         }
       }
     }
@@ -527,7 +516,7 @@ class ColMaker: public TreeUpdater {
                                const MetaInfo &info,
                                std::vector<ThreadEntry> &temp) { // NOLINT(*)
       // use cacheline aware optimization
-      if (TStats::kSimpleStats != 0 && param_.cache_opt != 0) {
+      if (GradStats::kSimpleStats != 0 && param_.cache_opt != 0) {
         EnumerateSplitCacheOpt(begin, end, d_step, fid, gpair, temp);
         return;
       }
@@ -537,7 +526,7 @@ class ColMaker: public TreeUpdater {
         temp[nid].stats.Clear();
       }
       // left statistics
-      TStats c(param_);
+      GradStats c(param_);
       for (const Entry *it = begin; it != end; it += d_step) {
         const bst_uint ridx = it->index;
         const int nid = position_[ridx];
@@ -559,16 +548,17 @@ class ColMaker: public TreeUpdater {
               bst_float loss_chg;
               if (d_step == -1) {
                 loss_chg = static_cast<bst_float>(
-                    constraints_[nid].CalcSplitGain(
-                        param_, param_.monotone_constraints[fid], c, e.stats) -
+                    spliteval_->ComputeSplitScore(nid, fid, c, e.stats) -
                     snode_[nid].root_gain);
+                e.best.Update(loss_chg, fid, (fvalue + e.last_fvalue) * 0.5f,
+                              d_step == -1, c, e.stats);
               } else {
                 loss_chg = static_cast<bst_float>(
-                    constraints_[nid].CalcSplitGain(
-                        param_, param_.monotone_constraints[fid], e.stats, c) -
+                    spliteval_->ComputeSplitScore(nid, fid, e.stats, c) -
                     snode_[nid].root_gain);
+                e.best.Update(loss_chg, fid, (fvalue + e.last_fvalue) * 0.5f,
+                              d_step == -1, e.stats, c);
               }
-              e.best.Update(loss_chg, fid, (fvalue + e.last_fvalue) * 0.5f, d_step == -1);
             }
           }
           // update the statistics
@@ -583,30 +573,31 @@ class ColMaker: public TreeUpdater {
         if (e.stats.sum_hess >= param_.min_child_weight &&
             c.sum_hess >= param_.min_child_weight) {
           bst_float loss_chg;
+          GradStats left_sum;
+          GradStats right_sum;
           if (d_step == -1) {
-            loss_chg = static_cast<bst_float>(
-                constraints_[nid].CalcSplitGain(
-                    param_, param_.monotone_constraints[fid], c, e.stats) -
-                snode_[nid].root_gain);
+            left_sum = c;
+            right_sum = e.stats;
           } else {
-            loss_chg = static_cast<bst_float>(
-                constraints_[nid].CalcSplitGain(
-                    param_, param_.monotone_constraints[fid], e.stats, c) -
-                snode_[nid].root_gain);
+            left_sum = e.stats;
+            right_sum = c;
           }
+          loss_chg = static_cast<bst_float>(
+              spliteval_->ComputeSplitScore(nid, fid, left_sum, right_sum) -
+              snode_[nid].root_gain);
           const bst_float gap = std::abs(e.last_fvalue) + kRtEps;
           const bst_float delta = d_step == +1 ? gap: -gap;
-          e.best.Update(loss_chg, fid, e.last_fvalue + delta, d_step == -1);
+          e.best.Update(loss_chg, fid, e.last_fvalue + delta, d_step == -1, left_sum, right_sum);
         }
       }
     }
 
     // update the solution candidate
     virtual void UpdateSolution(const SparsePage &batch,
-                                const std::vector<bst_uint> &feat_set,
+                                const std::vector<int> &feat_set,
                                 const std::vector<GradientPair> &gpair,
-                                const DMatrix &fmat) {
-      const MetaInfo& info = fmat.Info();
+                                DMatrix*p_fmat) {
+      const MetaInfo& info = p_fmat->Info();
       // start enumeration
       const auto num_features = static_cast<bst_omp_uint>(feat_set.size());
       #if defined(_OPENMP)
@@ -622,20 +613,20 @@ class ColMaker: public TreeUpdater {
           int fid = feat_set[i];
           const int tid = omp_get_thread_num();
           auto c = batch[fid];
-          const bool ind = c.length != 0 && c.data[0].fvalue == c.data[c.length - 1].fvalue;
-          if (param_.NeedForwardSearch(fmat.GetColDensity(fid), ind)) {
-            this->EnumerateSplit(c.data, c.data + c.length, +1,
+          const bool ind = c.size() != 0 && c[0].fvalue == c[c.size() - 1].fvalue;
+          if (param_.NeedForwardSearch(p_fmat->GetColDensity(fid), ind)) {
+            this->EnumerateSplit(c.data(), c.data() + c.size(), +1,
                                  fid, gpair, info, stemp_[tid]);
           }
-          if (param_.NeedBackwardSearch(fmat.GetColDensity(fid), ind)) {
-            this->EnumerateSplit(c.data + c.length - 1, c.data - 1, -1,
+          if (param_.NeedBackwardSearch(p_fmat->GetColDensity(fid), ind)) {
+            this->EnumerateSplit(c.data() + c.size() - 1, c.data() - 1, -1,
                                  fid, gpair, info, stemp_[tid]);
           }
         }
       } else {
         for (bst_omp_uint fid = 0; fid < num_features; ++fid) {
           this->ParallelFindSplit(batch[fid], fid,
-                                  fmat, gpair);
+                                  p_fmat, gpair);
         }
       }
     }
@@ -645,18 +636,10 @@ class ColMaker: public TreeUpdater {
                           const std::vector<GradientPair> &gpair,
                           DMatrix *p_fmat,
                           RegTree *p_tree) {
-      std::vector<bst_uint> feat_set = feat_index_;
-      if (param_.colsample_bylevel != 1.0f) {
-        std::shuffle(feat_set.begin(), feat_set.end(), common::GlobalRandom());
-        unsigned n = std::max(static_cast<unsigned>(1),
-                              static_cast<unsigned>(param_.colsample_bylevel * feat_index_.size()));
-        CHECK_GT(param_.colsample_bylevel, 0U)
-            << "colsample_bylevel cannot be zero.";
-        feat_set.resize(n);
-      }
-      auto iter = p_fmat->ColIterator();
-      while (iter->Next()) {
-        this->UpdateSolution(iter->Value(), feat_set, gpair, *p_fmat);
+      auto p_feature_set = column_sampler_.GetFeatureSet(depth);
+      const auto& feat_set = *p_feature_set;
+      for (const auto &batch : p_fmat->GetSortedColumnBatches()) {
+        this->UpdateSolution(batch, feat_set, gpair, p_fmat);
       }
       // after this each thread's stemp will get the best candidates, aggregate results
       this->SyncBestSolution(qexpand);
@@ -665,11 +648,16 @@ class ColMaker: public TreeUpdater {
         NodeEntry &e = snode_[nid];
         // now we know the solution in snode[nid], set split
         if (e.best.loss_chg > kRtEps) {
-          p_tree->AddChilds(nid);
-          (*p_tree)[nid].SetSplit(e.best.SplitIndex(), e.best.split_value, e.best.DefaultLeft());
-          // mark right child as 0, to indicate fresh leaf
-          (*p_tree)[(*p_tree)[nid].LeftChild()].SetLeaf(0.0f, 0);
-          (*p_tree)[(*p_tree)[nid].RightChild()].SetLeaf(0.0f, 0);
+          bst_float left_leaf_weight =
+              spliteval_->ComputeWeight(nid, e.best.left_sum) *
+              param_.learning_rate;
+          bst_float right_leaf_weight =
+              spliteval_->ComputeWeight(nid, e.best.right_sum) *
+              param_.learning_rate;
+          p_tree->ExpandNode(nid, e.best.SplitIndex(), e.best.split_value,
+                             e.best.DefaultLeft(), e.weight, left_leaf_weight,
+                             right_leaf_weight, e.best.loss_chg,
+                             e.stats.sum_hess);
         } else {
           (*p_tree)[nid].SetLeaf(e.weight * param_.learning_rate);
         }
@@ -682,15 +670,13 @@ class ColMaker: public TreeUpdater {
       // set the positions in the nondefault
       this->SetNonDefaultPosition(qexpand, p_fmat, tree);
       // set rest of instances to default position
-      const RowSet &rowset = p_fmat->BufferedRowset();
       // set default direct nodes to default
       // for leaf nodes that are not fresh, mark then to ~nid,
       // so that they are ignored in future statistics collection
-      const auto ndata = static_cast<bst_omp_uint>(rowset.Size());
+      const auto ndata = static_cast<bst_omp_uint>(p_fmat->Info().num_row_);
 
       #pragma omp parallel for schedule(static)
-      for (bst_omp_uint i = 0; i < ndata; ++i) {
-        const bst_uint ridx = rowset[i];
+      for (bst_omp_uint ridx = 0; ridx < ndata; ++ridx) {
         CHECK_LT(ridx, position_.size())
             << "ridx exceed bound " << "ridx="<<  ridx << " pos=" << position_.size();
         const int nid = this->DecodePosition(ridx);
@@ -731,12 +717,10 @@ class ColMaker: public TreeUpdater {
       }
       std::sort(fsplits.begin(), fsplits.end());
       fsplits.resize(std::unique(fsplits.begin(), fsplits.end()) - fsplits.begin());
-      auto iter = p_fmat->ColIterator();
-      while (iter->Next()) {
-        auto batch = iter->Value();
+      for (const auto &batch : p_fmat->GetSortedColumnBatches()) {
         for (auto fid : fsplits) {
           auto col = batch[fid];
-          const auto ndata = static_cast<bst_omp_uint>(col.length);
+          const auto ndata = static_cast<bst_omp_uint>(col.size());
           #pragma omp parallel for schedule(static)
           for (bst_omp_uint j = 0; j < ndata; ++j) {
             const bst_uint ridx = col[j].index;
@@ -772,8 +756,7 @@ class ColMaker: public TreeUpdater {
     const TrainParam& param_;
     // number of omp thread used during training
     const int nthread_;
-    // Per feature: shuffle index of each feature index
-    std::vector<bst_uint> feat_index_;
+    common::ColumnSampler column_sampler_;
     // Instance Data: current node position in the tree of each instance
     std::vector<int> position_;
     // PerThread x PerTreeNode: statistics for per thread construction
@@ -782,46 +765,47 @@ class ColMaker: public TreeUpdater {
     std::vector<NodeEntry> snode_;
     /*! \brief queue of nodes to be expanded */
     std::vector<int> qexpand_;
-    // constraint value
-    std::vector<TConstraint> constraints_;
+    // Evaluates splits and computes optimal weights for a given split
+    std::unique_ptr<SplitEvaluator> spliteval_;
   };
 };
 
 // distributed column maker
-template<typename TStats, typename TConstraint>
-class DistColMaker : public ColMaker<TStats, TConstraint> {
+class DistColMaker : public ColMaker {
  public:
-  DistColMaker() : builder_(param_) {
-    pruner_.reset(TreeUpdater::Create("prune"));
-  }
   void Init(const std::vector<std::pair<std::string, std::string> >& args) override {
     param_.InitAllowUnknown(args);
+    pruner_.reset(TreeUpdater::Create("prune"));
     pruner_->Init(args);
+    spliteval_.reset(SplitEvaluator::Create(param_.split_evaluator));
+    spliteval_->Init(args);
   }
   void Update(HostDeviceVector<GradientPair> *gpair,
               DMatrix* dmat,
               const std::vector<RegTree*> &trees) override {
-    TStats::CheckInfo(dmat->Info());
+    GradStats::CheckInfo(dmat->Info());
     CHECK_EQ(trees.size(), 1U) << "DistColMaker: only support one tree at a time";
+    Builder builder(
+      param_,
+      std::unique_ptr<SplitEvaluator>(spliteval_->GetHostClone()));
     // build the tree
-    builder_.Update(gpair->HostVector(), dmat, trees[0]);
+    builder.Update(gpair->ConstHostVector(), dmat, trees[0]);
     //// prune the tree, note that pruner will sync the tree
     pruner_->Update(gpair, dmat, trees);
     // update position after the tree is pruned
-    builder_.UpdatePosition(dmat, *trees[0]);
+    builder.UpdatePosition(dmat, *trees[0]);
   }
 
  private:
-  class Builder : public ColMaker<TStats, TConstraint>::Builder {
+  class Builder : public ColMaker::Builder {
    public:
-    explicit Builder(const TrainParam &param)
-        : ColMaker<TStats, TConstraint>::Builder(param) {}
+    explicit Builder(const TrainParam &param,
+                     std::unique_ptr<SplitEvaluator> spliteval)
+        : ColMaker::Builder(param, std::move(spliteval)) {}
     inline void UpdatePosition(DMatrix* p_fmat, const RegTree &tree) {
-      const RowSet &rowset = p_fmat->BufferedRowset();
-      const auto ndata = static_cast<bst_omp_uint>(rowset.Size());
+      const auto ndata = static_cast<bst_omp_uint>(p_fmat->Info().num_row_);
       #pragma omp parallel for schedule(static)
-      for (bst_omp_uint i = 0; i < ndata; ++i) {
-        const bst_uint ridx = rowset[i];
+      for (bst_omp_uint ridx = 0; ridx < ndata; ++ridx) {
         int nid = this->DecodePosition(ridx);
         while (tree[nid].IsDeleted()) {
           nid = tree[nid].Parent();
@@ -859,12 +843,10 @@ class DistColMaker : public ColMaker<TStats, TConstraint> {
             boolmap_[j] = 0;
         }
       }
-      auto iter = p_fmat->ColIterator();
-      while (iter->Next()) {
-        auto batch = iter->Value();
+      for (const auto &batch : p_fmat->GetSortedColumnBatches()) {
         for (auto fid : fsplits) {
           auto col = batch[fid];
-          const auto ndata = static_cast<bst_omp_uint>(col.length);
+          const auto ndata = static_cast<bst_omp_uint>(col.size());
           #pragma omp parallel for schedule(static)
           for (bst_omp_uint j = 0; j < ndata; ++j) {
             const bst_uint ridx = col[j].index;
@@ -884,12 +866,10 @@ class DistColMaker : public ColMaker<TStats, TConstraint> {
       bitmap_.InitFromBool(boolmap_);
       // communicate bitmap
       rabit::Allreduce<rabit::op::BitOR>(dmlc::BeginPtr(bitmap_.data), bitmap_.data.size());
-      const RowSet &rowset = p_fmat->BufferedRowset();
       // get the new position
-      const auto ndata = static_cast<bst_omp_uint>(rowset.Size());
+      const auto ndata = static_cast<bst_omp_uint>(p_fmat->Info().num_row_);
       #pragma omp parallel for schedule(static)
-      for (bst_omp_uint i = 0; i < ndata; ++i) {
-        const bst_uint ridx = rowset[i];
+      for (bst_omp_uint ridx = 0; ridx < ndata; ++ridx) {
         const int nid = this->DecodePosition(ridx);
         if (bitmap_.Get(ridx)) {
           CHECK(!tree[nid].IsLeaf()) << "inconsistent reduce information";
@@ -929,55 +909,20 @@ class DistColMaker : public ColMaker<TStats, TConstraint> {
   std::unique_ptr<TreeUpdater> pruner_;
   // training parameter
   TrainParam param_;
-  // pointer to the builder
-  Builder builder_;
-};
-
-// simple switch to defer implementation.
-class TreeUpdaterSwitch : public TreeUpdater {
- public:
-  TreeUpdaterSwitch()  = default;
-  void Init(const std::vector<std::pair<std::string, std::string> >& args) override {
-    for (auto &kv : args) {
-      if (kv.first == "monotone_constraints" && kv.second.length() != 0) {
-        monotone_ = true;
-      }
-    }
-    if (inner_ == nullptr) {
-      if (monotone_) {
-        inner_.reset(new ColMaker<GradStats, ValueConstraint>());
-      } else {
-        inner_.reset(new ColMaker<GradStats, NoConstraint>());
-      }
-    }
-
-    inner_->Init(args);
-  }
-
-  void Update(HostDeviceVector<GradientPair>* gpair,
-              DMatrix* data,
-              const std::vector<RegTree*>& trees) override {
-    CHECK(inner_ != nullptr);
-    inner_->Update(gpair, data, trees);
-  }
-
- private:
-  //  monotone constraints
-  bool monotone_{false};
-  // internal implementation
-  std::unique_ptr<TreeUpdater> inner_;
+  // Cloned for each builder instantiation
+  std::unique_ptr<SplitEvaluator> spliteval_;
 };
 
 XGBOOST_REGISTER_TREE_UPDATER(ColMaker, "grow_colmaker")
 .describe("Grow tree with parallelization over columns.")
 .set_body([]() {
-    return new TreeUpdaterSwitch();
+    return new ColMaker();
   });
 
 XGBOOST_REGISTER_TREE_UPDATER(DistColMaker, "distcol")
 .describe("Distributed column split version of tree maker.")
 .set_body([]() {
-    return new DistColMaker<GradStats, NoConstraint>();
+    return new DistColMaker();
   });
 }  // namespace tree
 }  // namespace xgboost
