@@ -4,12 +4,13 @@
  * \brief refresh the statistics and leaf value on the tree on the dataset
  * \author Tianqi Chen
  */
-
+#include <rabit/rabit.h>
 #include <xgboost/tree_updater.h>
+
 #include <vector>
 #include <limits>
+
 #include "./param.h"
-#include "../common/sync.h"
 #include "../common/io.h"
 
 namespace xgboost {
@@ -29,7 +30,7 @@ class TreeRefresher: public TreeUpdater {
               DMatrix *p_fmat,
               const std::vector<RegTree*> &trees) override {
     if (trees.size() == 0) return;
-    std::vector<GradientPair> &gpair_h = gpair->HostVector();
+    const std::vector<GradientPair> &gpair_h = gpair->ConstHostVector();
     // number of threads
     // thread temporal space
     std::vector<std::vector<TStats> > stemp;
@@ -57,10 +58,7 @@ class TreeRefresher: public TreeUpdater {
     {
       const MetaInfo &info = p_fmat->Info();
       // start accumulating statistics
-       auto *iter = p_fmat->RowIterator();
-      iter->BeforeFirst();
-      while (iter->Next()) {
-         auto batch = iter->Value();
+      for (const auto &batch : p_fmat->GetRowBatches()) {
         CHECK_LT(batch.Size(), std::numeric_limits<unsigned>::max());
         const auto nbatch = static_cast<bst_omp_uint>(batch.Size());
         #pragma omp parallel for schedule(static)
@@ -129,7 +127,6 @@ class TreeRefresher: public TreeUpdater {
     RegTree &tree = *p_tree;
     tree.Stat(nid).base_weight = static_cast<bst_float>(gstats[nid].CalcWeight(param_));
     tree.Stat(nid).sum_hess = static_cast<bst_float>(gstats[nid].sum_hess);
-    gstats[nid].SetLeafVec(param_, tree.Leafvec(nid));
     if (tree[nid].IsLeaf()) {
       if (param_.refresh_leaf) {
         tree[nid].SetLeaf(tree.Stat(nid).base_weight * param_.learning_rate);
