@@ -5,12 +5,12 @@
           a refresh is needed to make the statistics exactly correct
  * \author Tianqi Chen
  */
-
+#include <rabit/rabit.h>
 #include <xgboost/base.h>
 #include <xgboost/tree_updater.h>
 #include <vector>
 #include <algorithm>
-#include "../common/sync.h"
+
 #include "../common/quantile.h"
 #include "../common/group_data.h"
 #include "./updater_basemaker-inl.h"
@@ -30,7 +30,7 @@ class SketchMaker: public BaseMaker {
     param_.learning_rate = lr / trees.size();
     // build tree
     for (auto tree : trees) {
-      this->Update(gpair->HostVector(), p_fmat, tree);
+      this->Update(gpair->ConstHostVector(), p_fmat, tree);
     }
     param_.learning_rate = lr;
   }
@@ -128,9 +128,6 @@ class SketchMaker: public BaseMaker {
     inline static void Reduce(SKStats &a, const SKStats &b) { // NOLINT(*)
       a.Add(b);
     }
-    /*! \brief set leaf vector value based on statistics */
-    inline void SetLeafVec(const TrainParam &param, bst_float *vec) const {
-    }
   };
   inline void BuildSketch(const std::vector<GradientPair> &gpair,
                           DMatrix *p_fmat,
@@ -142,12 +139,9 @@ class SketchMaker: public BaseMaker {
     }
     thread_sketch_.resize(omp_get_max_threads());
     // number of rows in
-    const size_t nrows = p_fmat->BufferedRowset().Size();
+    const size_t nrows = p_fmat->Info().num_row_;
     // start accumulating statistics
-    auto iter = p_fmat->ColIterator();
-    iter->BeforeFirst();
-    while (iter->Next()) {
-      auto batch = iter->Value();
+    for (const auto &batch : p_fmat->GetSortedColumnBatches()) {
       // start enumeration
       const auto nsize = static_cast<bst_omp_uint>(batch.Size());
       #pragma omp parallel for schedule(dynamic, 1)
@@ -155,7 +149,7 @@ class SketchMaker: public BaseMaker {
         this->UpdateSketchCol(gpair, batch[fidx], tree,
                               node_stats_,
                               fidx,
-                              batch[fidx].length == nrows,
+                              batch[fidx].size() == nrows,
                               &thread_sketch_[omp_get_thread_num()]);
       }
     }
@@ -174,13 +168,13 @@ class SketchMaker: public BaseMaker {
   }
   // update sketch information in column fid
   inline void UpdateSketchCol(const std::vector<GradientPair> &gpair,
-                              const SparsePage::Inst &c,
+                              const SparsePage::Inst &col,
                               const RegTree &tree,
                               const std::vector<SKStats> &nstats,
                               bst_uint fid,
                               bool col_full,
                               std::vector<SketchEntry> *p_temp) {
-    if (c.length == 0) return;
+    if (col.size() == 0) return;
     // initialize sbuilder for use
     std::vector<SketchEntry> &sbuilder = *p_temp;
     sbuilder.resize(tree.param.num_nodes * 3);
@@ -192,10 +186,10 @@ class SketchMaker: public BaseMaker {
       }
     }
     if (!col_full) {
-      for (bst_uint j = 0; j < c.length; ++j) {
-        const bst_uint ridx = c[j].index;
+      for (const auto& c : col) {
+        const bst_uint ridx = c.index;
         const int nid = this->position_[ridx];
-        if (nid >= 0) {
+        if (nid > 0) {
           const GradientPair &e = gpair[ridx];
           if (e.GetGrad() >= 0.0f) {
             sbuilder[3 * nid + 0].sum_total += e.GetGrad();
@@ -213,10 +207,10 @@ class SketchMaker: public BaseMaker {
       }
     }
     // if only one value, no need to do second pass
-    if (c[0].fvalue  == c[c.length-1].fvalue) {
+    if (col[0].fvalue  == col[col.size()-1].fvalue) {
       for (int nid : this->qexpand_) {
         for (int k = 0; k < 3; ++k) {
-          sbuilder[3 * nid + k].sketch->Push(c[0].fvalue,
+          sbuilder[3 * nid + k].sketch->Push(col[0].fvalue,
                                              static_cast<bst_float>(
                                                  sbuilder[3 * nid + k].sum_total));
         }
@@ -231,17 +225,17 @@ class SketchMaker: public BaseMaker {
       }
     }
     // second pass, build the sketch
-    for (bst_uint j = 0; j < c.length; ++j) {
-      const bst_uint ridx = c[j].index;
+    for (const auto& c : col) {
+      const bst_uint ridx = c.index;
       const int nid = this->position_[ridx];
       if (nid >= 0) {
         const GradientPair &e = gpair[ridx];
         if (e.GetGrad() >= 0.0f) {
-          sbuilder[3 * nid + 0].Push(c[j].fvalue, e.GetGrad(), max_size);
+          sbuilder[3 * nid + 0].Push(c.fvalue, e.GetGrad(), max_size);
         } else {
-          sbuilder[3 * nid + 1].Push(c[j].fvalue, -e.GetGrad(), max_size);
+          sbuilder[3 * nid + 1].Push(c.fvalue, -e.GetGrad(), max_size);
         }
-        sbuilder[3 * nid + 2].Push(c[j].fvalue, e.GetHess(), max_size);
+        sbuilder[3 * nid + 2].Push(c.fvalue, e.GetHess(), max_size);
       }
     }
     for (int nid : this->qexpand_) {
@@ -287,16 +281,21 @@ class SketchMaker: public BaseMaker {
       const int nid = qexpand_[wid];
       const SplitEntry &best = sol[wid];
       // set up the values
-      p_tree->Stat(nid).loss_chg = best.loss_chg;
       this->SetStats(nid, node_stats_[nid], p_tree);
       // now we know the solution in snode[nid], set split
       if (best.loss_chg > kRtEps) {
-        p_tree->AddChilds(nid);
-        (*p_tree)[nid].SetSplit(best.SplitIndex(),
-                                 best.split_value, best.DefaultLeft());
-        // mark right child as 0, to indicate fresh leaf
-        (*p_tree)[(*p_tree)[nid].LeftChild()].SetLeaf(0.0f, 0);
-        (*p_tree)[(*p_tree)[nid].RightChild()].SetLeaf(0.0f, 0);
+        bst_float base_weight = node_stats_[nid].CalcWeight(param_);
+        bst_float left_leaf_weight =
+            CalcWeight(param_, best.left_sum.sum_grad, best.left_sum.sum_hess) *
+            param_.learning_rate;
+        bst_float right_leaf_weight =
+            CalcWeight(param_, best.right_sum.sum_grad,
+                       best.right_sum.sum_hess) *
+            param_.learning_rate;
+        p_tree->ExpandNode(nid, best.SplitIndex(), best.split_value,
+                           best.DefaultLeft(), base_weight, left_leaf_weight,
+                           right_leaf_weight, best.loss_chg,
+                           node_stats_[nid].sum_hess);
       } else {
         (*p_tree)[nid].SetLeaf(p_tree->Stat(nid).base_weight * param_.learning_rate);
       }
@@ -306,7 +305,6 @@ class SketchMaker: public BaseMaker {
   inline void SetStats(int nid, const SKStats &node_sum, RegTree *p_tree) {
     p_tree->Stat(nid).base_weight = static_cast<bst_float>(node_sum.CalcWeight(param_));
     p_tree->Stat(nid).sum_hess = static_cast<bst_float>(node_sum.sum_hess);
-    node_sum.SetLeafVec(param_, p_tree->Leafvec(nid));
   }
   inline void EnumerateSplit(const WXQSketch::Summary &pos_grad,
                              const WXQSketch::Summary &neg_grad,
@@ -347,7 +345,9 @@ class SketchMaker: public BaseMaker {
       if (s.sum_hess >= param_.min_child_weight &&
           c.sum_hess >= param_.min_child_weight) {
         double loss_chg = s.CalcGain(param_) + c.CalcGain(param_) - root_gain;
-        best->Update(static_cast<bst_float>(loss_chg), fid, fsplits[i], false);
+        best->Update(static_cast<bst_float>(loss_chg), fid, fsplits[i], false,
+                     GradStats(s.pos_grad - s.neg_grad , s.sum_hess),
+                     GradStats(c.pos_grad - c.neg_grad, c.sum_hess));
       }
       // backward
       c.SetSubstract(feat_sum, s);
@@ -355,7 +355,9 @@ class SketchMaker: public BaseMaker {
       if (s.sum_hess >= param_.min_child_weight &&
           c.sum_hess >= param_.min_child_weight) {
         double loss_chg = s.CalcGain(param_) + c.CalcGain(param_) - root_gain;
-        best->Update(static_cast<bst_float>(loss_chg), fid, fsplits[i], true);
+        best->Update(static_cast<bst_float>(loss_chg), fid, fsplits[i], true,
+                     GradStats(s.pos_grad - s.neg_grad, s.sum_hess),
+                     GradStats(c.pos_grad - c.neg_grad, c.sum_hess));
       }
     }
     {
@@ -366,8 +368,10 @@ class SketchMaker: public BaseMaker {
           c.sum_hess >= param_.min_child_weight) {
         bst_float cpt = fsplits.back();
         double loss_chg = s.CalcGain(param_) + c.CalcGain(param_) - root_gain;
-        best->Update(static_cast<bst_float>(loss_chg),
-                     fid, cpt + std::abs(cpt) + 1.0f, false);
+        best->Update(static_cast<bst_float>(loss_chg), fid,
+                     cpt + std::abs(cpt) + 1.0f, false,
+                     GradStats(s.pos_grad - s.neg_grad, s.sum_hess),
+                     GradStats(c.pos_grad - c.neg_grad, c.sum_hess));
       }
     }
   }
